import json
from pathlib import Path

import pytest


@pytest.fixture
def test_data_dir():
    return Path(__file__).parent / "test_data"


@pytest.fixture
def schema_file_path(test_data_dir):
    return test_data_dir / "schema.json"


@pytest.fixture
def schema_dict(schema_file_path):
    return json.loads(schema_file_path.read_text())



"""Exceptions for working with JSON Schema specifications."""


class SchemaError(Exception):
    """Generic JSON Schema error."""


class SchemaParseError(SchemaError):
    """Error while parsing JSON Schema."""


class DataValidationError(SchemaError):
    """Error on validation of data."""



from datetime import datetime

from momoa import Schema
from momoa.model import UNDEFINED


def test_instantiated_object_converts_to_serialized_data_with_validation(schema_dict):
    schema = Schema(schema_dict)
    data_object = schema.model(
        firstName="Boris",
        lastName="Harrison",
        age=53,
        deceased=False,
        birthday=datetime(1969, 11, 23),
    )

    result = data_object.serialize()

    assert result == {
        "firstName": "Boris",
        "lastName": "Harrison",
        "age": 53,
        "gender": "male",
        "deceased": False,
        "birthday": "1969-11-23",
    }


def test_nested_models_are_serialized(schema_dict):
    schema = Schema(schema_dict)
    data_object = schema.model(
        firstName="Boris",
        lastName="Harrison",
        age=53,
        deceased=False,
        birthday=datetime(1969, 11, 23),
        address={
            "street": "foo",
            "city": "bar",
        },
    )

    result = data_object.serialize()

    assert result == {
        "firstName": "Boris",
        "lastName": "Harrison",
        "age": 53,
        "gender": "male",
        "deceased": False,
        "birthday": "1969-11-23",
        "address": {
            "street": "foo",
            "city": "bar",
        },
    }


def test_attributes_set_after_instantiation_serialize_correctly(schema_dict):
    schema = Schema(schema_dict)
    person = schema.model(firstName="Boris", lastName="Harrison")

    assert person.age is UNDEFINED
    assert person.birthday is UNDEFINED

    serialized = person.serialize()

    assert "age" not in serialized
    assert "birthday" not in serialized

    person.age = 53
    person.birthday = datetime(1969, 11, 23)

    serialized = person.serialize()

    assert serialized["age"] == 53
    assert serialized["birthday"] == "1969-11-23"


def test_generic_subschemas_are_serialized_correctly(test_data_dir):
    description_schema = Schema.from_file(test_data_dir / "action_description.json")
    input_schema = Schema.from_file(test_data_dir / "action_input_schema.json")
    output_schema = Schema.from_file(test_data_dir / "action_output_schema.json")
    description = {
        "action": "test_action",
        "input_schema": input_schema.schema_dict,
        "output_schema": output_schema.schema_dict,
    }

    description_model = description_schema.model(**description)

    serialized = description_model.serialize()
    assert serialized == {
        "action": "test_action",
        "input_schema": {
            "type": "object",
            "title": "Get Example Input Schema",
            "properties": {"something": {"type": "string"}},
            "required": ["something"],
        },
        "output_schema": {
            "type": "object",
            "title": "Get Example Output Schema",
            "properties": {"something_else": {"type": "string"}},
            "required": ["something_else"],
        },
    }


def test_serialization_is_inverse_of_deserialization_if_no_undefined_defaults(schema_dict):
    """Note: Works only if there are no undefined values for properties with a default."""
    schema = Schema(schema_dict)

    test_data = {
        "firstName": "Boris",
        "lastName": "Harrison",
        "age": 53,
        "dogs": ["Fluffy", "Crumpet"],
        "gender": "other",
        "deceased": False,
        "address": {
            "street": "adipisicing do proident laborum",
            "city": "veniam nulla ipsum adipisicing eu",
            "state": "Excepteur esse elit",
        },
    }
    deserialized = schema.deserialize(test_data)
    serialized = deserialized.serialize()

    assert serialized == test_data



"""Basic class to parse a schema and prepare the model class."""

from __future__ import annotations

import json
from copy import deepcopy
from functools import cached_property
from pathlib import Path
from typing import Any, Dict, Mapping, Sequence, Type, Union

from json_ref_dict import materialize, RefDict
from statham.schema.parser import parse
from statham.serializers.orderer import orderer
from statham.titles import title_labeller

from .exceptions import SchemaParseError
from .model import Model, ModelFactory


class Schema:
    """Basic class to parse the schema and prepare the model class."""

    def __init__(self, schema: Dict[str, Any], model_factory: ModelFactory = Model.make_model):
        """
        Constructs the Schema class instance.

        Arguments:
            schema: A Python dict representation of the JSONSchema specification.
            model_factory: A callable that creates a model subclass from a JSON Schema.
                           Can be used to customize Model creation.
        """
        self.schema_dict = schema
        self.title: str = self.schema_dict["title"]
        try:
            parsed = parse(deepcopy(self.schema_dict))
        except KeyError as ex:
            raise SchemaParseError(f"Error parsing schema `{self.title}`: {ex}") from ex
        else:
            self.models: Sequence[Type[Model]] = tuple(map(model_factory, orderer(*parsed)))

    @classmethod
    def from_uri(cls, input_uri: str) -> Schema:
        """
        Instantiates the Schema from a URI to the schema document.

        For local files use the `file://` scheme. This method also dereferences
        the internal `$ref` links.

        Arguments:
            input_uri: String representation of the URI to the schema.

        Returns:
            Schema instance.
        """
        return cls(materialize(RefDict.from_uri(input_uri), context_labeller=title_labeller()))

    @classmethod
    def from_file(cls, file_path: Union[Path, str]) -> Schema:
        """
        Helper to instantiate the Schema from a local file path.

        Note: This method will _not_ dereference any internal `$ref` links.

        Arguments:
            file_path: Either a simple string path or a `pathlib.Path` object.

        Returns:
            Schema instance.
        """
        return cls.from_uri(Path(file_path).absolute().as_uri())

    @cached_property
    def model(self) -> Type[Model]:
        """
        Retrieves the top model class of the schema.

        Returns:
            Model subclass.
        """
        return self.models[-1]

    def deserialize(self, raw_data: Union[Mapping[str, Any], str]) -> Model:
        """
        Converts raw data to the Model instance, validating it in the process.

        Arguments:
            raw_data: The raw data to deserialize. Can be either a JSON string
                or a preloaded Python mapping object.

        Returns:
            An instance of the Model class.
        """
        if isinstance(raw_data, str):
            raw_data = json.loads(raw_data)
        return self.model(**raw_data)  # type: ignore



import inspect
from datetime import datetime

import pytest
from statham.schema.parser import parse

from momoa.exceptions import DataValidationError
from momoa.model import Model, UNDEFINED


@pytest.fixture
def PersonModel(schema_dict):
    schema_class = parse(schema_dict).pop()
    return Model.make_model(schema_class)


def test_make_model_creates_model_class(schema_dict):
    schema_class = parse(schema_dict).pop()
    model = Model.make_model(schema_class)

    assert inspect.isclass(model)
    assert issubclass(model, Model)
    assert model.__name__ == "PersonModel"
    assert model._schema_class is schema_class


def test_instantiation_without_required_fields_raises_validation_error(PersonModel):
    with pytest.raises(DataValidationError):
        PersonModel()


def test_instantiation_with_required_values_also_sets_default_values(PersonModel):
    data_object = PersonModel(firstName="Boris", lastName="Harrison")
    assert data_object.firstName == "Boris"
    assert data_object.lastName == "Harrison"
    assert data_object.gender == "male"
    assert data_object.deceased is UNDEFINED


def test_instantiation_with_incorrect_values_raises_validation_error(PersonModel):
    with pytest.raises(DataValidationError):
        PersonModel(firstName="Boris", lastName="Harrison", age="53")


def test_instantiation_with_incorrect_fields_raises_validation_error(PersonModel):
    with pytest.raises(DataValidationError):
        PersonModel(firstName="Boris", lastName="Harrison", foo="bar")


def test_instantiation_with_native_datetime(PersonModel):
    birthday = datetime(1969, 11, 23)
    person = PersonModel(firstName="Boris", lastName="Harrison", birthday=birthday)

    assert person.birthday == birthday


def test_instantiation_with_iso_datetime(PersonModel):
    birthday = "1969-11-23"
    person = PersonModel(firstName="Boris", lastName="Harrison", birthday=birthday)

    assert person.birthday == datetime(1969, 11, 23)


def test_invalid_attribute_raises_exception(PersonModel):
    person = PersonModel(firstName="Boris", lastName="Harrison")
    with pytest.raises(AttributeError):
        person.foobar


def test_unset_attribute_returns_sentinel_value(PersonModel):
    person = PersonModel(firstName="Boris", lastName="Harrison")
    assert person.age is UNDEFINED


def test_setting_attributes_after_instantiation(PersonModel):
    birthday = datetime(1969, 11, 23)
    person = PersonModel(firstName="Boris", lastName="Harrison")

    assert person.age is UNDEFINED
    assert person.birthday is UNDEFINED

    person.age = 53
    person.birthday = birthday

    assert person.age == 53
    assert person.birthday == birthday


def test_model_is_iterable(PersonModel):
    person = PersonModel(firstName="Boris", lastName="Harrison")

    for field, value in person:
        assert value == getattr(person, field)


def test_models_with_same_data_are_equal_but_not_identical(PersonModel):
    person_1 = PersonModel(firstName="Boris", lastName="Harrison", dogs=["Fluffy", "Cerberus"])
    person_2 = PersonModel(firstName="Boris", lastName="Harrison", dogs=["Fluffy", "Cerberus"])

    assert person_1 == person_2
    assert person_1 is not person_2


def test_models_with_different_data_are_not_equal(PersonModel):
    person_1 = PersonModel(firstName="Boris", lastName="Harrison", dogs=["Fluffy", "Cerberus"])
    person_2 = PersonModel(firstName="Boris", lastName="Harrison", dogs=["Fluffy"])

    assert person_1 != person_2



import json
from datetime import datetime

import pytest
from statham.schema.constants import NotPassed

from momoa import Schema
from momoa.exceptions import DataValidationError
from momoa.model import UNDEFINED


def test_serialized_data_is_validated_and_converted_to_python_from_dict(schema_dict):
    test_data = {
        "firstName": "Boris",
        "lastName": "Harrison",
        "age": 53,
        "dogs": ["Fluffy", "Crumpet"],
        "gender": "male",
        "deceased": False,
        "address": {
            "street": "adipisicing do proident laborum",
            "city": "veniam nulla ipsum adipisicing eu",
            "state": "Excepteur esse elit",
        },
    }

    schema = Schema(schema_dict)
    result = schema.deserialize(test_data)

    assert type(result).__name__ == "PersonModel"
    assert isinstance(result, schema.model)
    assert result.firstName == "Boris"
    assert result.lastName == "Harrison"
    assert result.gender == "male"
    assert not result.deceased
    assert result.birthday is UNDEFINED


def test_serialized_data_is_validated_and_converted_to_python_from_string(schema_dict):
    test_data = json.dumps(
        {
            "firstName": "Boris",
            "lastName": "Harrison",
            "age": 53,
            "dogs": ["Fluffy", "Crumpet"],
            "gender": "male",
            "deceased": False,
            "address": {
                "street": "adipisicing do proident laborum",
                "city": "veniam nulla ipsum adipisicing eu",
                "state": "Excepteur esse elit",
            },
        }
    )

    schema = Schema(schema_dict)
    result = schema.deserialize(test_data)

    assert type(result).__name__ == "PersonModel"
    assert isinstance(result, schema.model)
    assert result.firstName == "Boris"
    assert result.lastName == "Harrison"
    assert result.birthday is NotPassed()
    assert not result.deceased


def test_invalid_data_raises_exception(schema_dict):
    test_data = json.dumps(
        {
            "firstName": "Boris",
            "lastName": "Harrison",
            "age": "53",
            "dogs": ["Fluffy", "Crumpet"],
            "gender": "male",
            "deceased": False,
            "address": {
                "street": "adipisicing do proident laborum",
                "city": "veniam nulla ipsum adipisicing eu",
                "state": "Excepteur esse elit",
            },
        }
    )

    schema = Schema(schema_dict)

    with pytest.raises(DataValidationError):
        schema.deserialize(test_data)


def test_serialized_data_containing_datetime_is_validated_and_converted(schema_dict):
    test_data = {"firstName": "Boris", "lastName": "Harrison", "birthday": "1969-11-23"}

    schema = Schema(schema_dict)
    result = schema.deserialize(test_data)

    assert type(result).__name__ == "PersonModel"
    assert isinstance(result, schema.model)
    assert result.birthday == datetime(1969, 11, 23)


def test_deserialization_is_inverse_of_serialization(schema_dict):
    schema = Schema(schema_dict)

    test_data = {
        "firstName": "Boris",
        "lastName": "Harrison",
        "age": 53,
        "dogs": ["Fluffy", "Crumpet"],
        "gender": "male",
        "deceased": False,
        "address": {
            "street": "adipisicing do proident laborum",
            "city": "veniam nulla ipsum adipisicing eu",
            "state": "Excepteur esse elit",
        },
    }

    instance = schema.model(**test_data)
    serialized = instance.serialize()
    deserialized = schema.deserialize(serialized)

    assert deserialized == instance


def test_deserialization_creates_default_values(schema_dict):
    schema = Schema(schema_dict)

    test_data = {
        "firstName": "Boris",
        "lastName": "Harrison",
        "age": 53,
        "dogs": ["Fluffy", "Crumpet"],
        "deceased": False,
        "address": {
            "street": "adipisicing do proident laborum",
            "city": "veniam nulla ipsum adipisicing eu",
            "state": "Excepteur esse elit",
        },
    }
    deserialized = schema.deserialize(test_data)

    assert deserialized.gender == "male"



"""Base wrapper class for building JSONSchema based models."""

from __future__ import annotations

from typing import Any, Callable, cast, Type

from humps import pascalize
from statham.schema.constants import NotPassed
from statham.schema.elements import meta, String
from statham.schema.exceptions import ValidationError

from .exceptions import DataValidationError
from .format import StringFormat

# Sentinel for unset values.
UNDEFINED = NotPassed()


class Model:
    """Base model class."""

    _schema_class: meta.ObjectMeta
    _formatter: Type[StringFormat]

    def __init__(self, **data):
        try:
            self._instance = self._schema_class(
                {key: self._format(key, value) for key, value in data.items()}
            )
        except ValidationError as ex:
            raise DataValidationError(f"{type(self).__name__} validation error: {ex}") from ex

    def _format(self, field: str, value: Any) -> str:
        """Converts Python native values to JSONSchema string equivalents on the fly."""
        element = self._get_field_element(field)
        if isinstance(element, String) and not isinstance(value, str):
            value = self._formatter(element.format).to_(value)
        return value

    def _unformat(self, field: str, value: str) -> Any:
        """Converts JSONSchema formatted string values to Python native on the fly."""
        element = self._get_field_element(field)
        if isinstance(element, String) and value:
            value = self._formatter(element.format).from_(value)
        else:
            value = element(value)
        return value

    def __getattr__(self, item: str) -> Any:
        if item in self._schema_class.properties:  # type: ignore
            return self._unformat(item, getattr(self._instance, item))
        raise AttributeError(f"'{type(self).__name__}' object has no attribute '{item}'")

    def __setattr__(self, item: str, value: Any) -> None:
        if item in self._schema_class.properties:  # type: ignore
            formatted_value = self._format(item, value)
            setattr(self._instance, item, formatted_value)
            self._instance._dict[item] = formatted_value
        else:
            super().__setattr__(item, value)

    def __iter__(self):
        return (
            (field_name, getattr(self, field_name))
            for field_name in self._schema_class.properties
        )

    def __eq__(self, other) -> bool:
        return isinstance(other, Model) and all(
            getattr(other, field) == value for field, value in self
        )

    def _get_field_element(self, field):
        try:
            return self._schema_class.properties[field].element
        except KeyError as ex:
            raise DataValidationError(f"Invalid field '{field}'") from ex

    def serialize(self):
        """Validates data and serializes it into JSON-ready format."""
        return _serialize_schema_value(self._instance)

    @staticmethod  # pragma: no mutate
    def make_model(schema_class: meta.ObjectMeta, string_formatter=StringFormat) -> Type[Model]:
        """
        Constructs a Model subclass based on the class derived from JSONSchema.

        Args:
            schema_class: Class derived from the JSONSchema.
            string_formatter: Class used to format strings.

        Returns:
            Subclass of the Model class.
        """
        name = pascalize(schema_class.__name__) + "Model"
        return cast(
            Type[Model],
            type(
                name, (Model,), {"_schema_class": schema_class, "_formatter": string_formatter}
            ),
        )


ModelFactory = Callable[[meta.ObjectMeta], Type[Model]]  # pragma: no mutate


def _serialize_schema_value(value: Any) -> Any:
    """Helper function to recursively serialize schema values."""
    if isinstance(value, list):
        return [_serialize_schema_value(item) for item in value]
    if isinstance(type(value), meta.ObjectMeta):
        value = value._dict
    if isinstance(value, dict):
        return {
            field_name: _serialize_schema_value(field_value)
            for field_name, field_value in value.items()
            if not field_name.startswith("_") and field_value is not UNDEFINED
        }
    return value



"""Helper utilities for formatting serialized data."""

from datetime import date, datetime
from ipaddress import ip_address
from typing import Any, Callable, cast, Mapping
from uuid import UUID

from dateutil.parser import isoparse

FormatMapping = Mapping[str, Callable]


def _format_date_time(value: datetime) -> str:
    """
    Helper function for formatting date-time values.

    If all the sub-day units are set to `00`, the value is formatted as a date.
    """
    if not any((value.hour, value.minute, value.second, value.microsecond)):
        value = cast(datetime, date(value.year, value.month, value.day))
    return value.isoformat()


class StringFormat:
    """
    Helper utility for converting values of string fields with format.

    JSON Schema `string` type supports the `format` keyword, which can be used to
    convert its value to and from a native Python type. This class supports custom
    implementation of that conversion.
    """

    _to_mapping: FormatMapping = {"date-time": _format_date_time}
    _from_mapping: FormatMapping = {
        "date-time": isoparse,
        "ipv4": ip_address,
        "ipv6": ip_address,
        "uuid": UUID,
    }

    def __init__(self, fmt):
        self.format = fmt

    def to_(self, value: Any) -> str:
        """
        Converts a value of a Python native type to a formatted JSON string.

        Args:
            value: Any Python type, to be converted if supported.

        Returns:
            A formatted string.
        """
        return self._to_mapping.get(self.format, str)(value)

    def from_(self, value: str) -> Any:
        """
        Converts a formatted JSON string to a native Python type.

        Args:
            value: A formatted string, to be converted if supported.

        Returns:
            A native Python type if conversion is supported, or the original
            string value if not.
        """
        return self._from_mapping.get(self.format, str)(value)



from datetime import datetime, timezone
from ipaddress import IPv4Address, IPv6Address
from uuid import UUID

import pytest

from momoa.format import StringFormat


@pytest.mark.parametrize(
    "format_name, value, deserialized",
    (
        pytest.param("date-time", "2002-05-22", datetime(2002, 5, 22), id="date"),
        pytest.param(
            "date-time",
            "2002-05-22T11:45:00",
            datetime(2002, 5, 22, 11, 45),
            id="datetime",
        ),
        pytest.param(
            "date-time",
            "2002-05-22T11:45:00Z",
            datetime(2002, 5, 22, 11, 45, tzinfo=timezone.utc),
            id="datetime with timezone",
        ),
        pytest.param("ipv4", "192.168.0.1", IPv4Address("192.168.0.1"), id="IPv4"),
        pytest.param("ipv6", "2001:db8::1000", IPv6Address("2001:db8::1000"), id="IPv6"),
        pytest.param(
            "uuid",
            "4cbf6ab8-1785-4264-a011-ee655edbe4f5",
            UUID("4cbf6ab8-1785-4264-a011-ee655edbe4f5"),
            id="UUID",
        ),
        pytest.param("unknown", "blabla", "blabla", id="unknown format"),
    ),
)
def test_strings_are_deserialised_to_correct_value(format_name, value, deserialized):
    fmt = StringFormat(format_name)
    assert fmt.from_(value) == deserialized


@pytest.mark.parametrize(
    "format_name, serialized, value",
    (
        pytest.param("date-time", "2002-05-22", datetime(2002, 5, 22), id="date"),
        pytest.param(
            "date-time",
            "2002-05-22T11:45:00",
            datetime(2002, 5, 22, 11, 45),
            id="datetime",
        ),
        pytest.param(
            "date-time",
            "2002-05-22T11:45:00+00:00",
            datetime(2002, 5, 22, 11, 45, tzinfo=timezone.utc),
            id="datetime with timezone",
        ),
        pytest.param("unknown", "blabla", "blabla", id="unknown format"),
        pytest.param(
            "uuid",
            "4cbf6ab8-1785-4264-a011-ee655edbe4f5",
            UUID("4cbf6ab8-1785-4264-a011-ee655edbe4f5"),
            id="UUID",
        ),
    ),
)
def test_values_are_serialized_to_correct_format(format_name, serialized, value):
    fmt = StringFormat(format_name)
    assert fmt.to_(value) == serialized



import logging

import pytest

from momoa import Schema
from momoa.exceptions import SchemaError
from momoa.model import Model


def test_valid_schema_loads_from_dict(schema_dict, caplog):
    caplog.set_level(logging.DEBUG)

    loaded_schema = Schema(schema_dict)

    assert loaded_schema.schema_dict == schema_dict
    assert loaded_schema.title == "Person"

    model_names = ("AddressModel", "ShoePreferencesModel", "PersonModel")
    assert loaded_schema.model == loaded_schema.models[-1]
    assert loaded_schema.model.__name__ == model_names[-1]

    for index, model in enumerate(loaded_schema.models):
        assert issubclass(model, Model)
        assert model.__name__ == model_names[index]


def test_invalid_schema_fails_loading(schema_dict):
    schema_dict["properties"]["lastName"]["type"] = "blabla"

    with pytest.raises(SchemaError):
        Schema(schema_dict)


def test_valid_schema_loads_from_uri(schema_file_path):
    loaded_schema = Schema.from_uri(f"file://{schema_file_path}")

    for model in loaded_schema.models:
        assert issubclass(model, Model)


def test_valid_schema_loads_from_file_path(schema_file_path):
    loaded_schema = Schema.from_file(schema_file_path)

    for model in loaded_schema.models:
        assert issubclass(model, Model)


def test_custom_model_factory_creates_model(schema_dict):
    def custom_model_factory(schema_class):
        name = schema_class.__name__.lower() + "blabla"
        return type(name, (Model,), {"_schema_class": schema_class})

    loaded_schema = Schema(schema_dict, model_factory=custom_model_factory)
    assert loaded_schema.model.__name__ == "personblabla"
